/*
 * File name: sndsimulator.cc
 * Date:      Mon Oct 15 14:50:45 +0200 2012
 * Author:    Miroslav Kulich
 */

#include <unistd.h>
#include <iostream>
#include <sstream>
#include <vtkImageData.h>
#include <math.h>

#include "lasersimulator.h"

using namespace imr;

/// Callbacks ------------------------------------------------------ 

/**
 * Function called by libfov to apply light to a cell.
 *
 * \param map Pointer to map data structure passed to function such as
 *            fov_circle.
 * \param x   Absolute x-axis position of cell.
 * \param y   Absolute x-axis position of cell.
 * \param dx  Offset of cell from source cell on x-axis.
 * \param dy  Offset of cell from source cell on y-axis.
 * \param src Pointer to source data structure passed to function such
 *            as fov_circle.
 */
void apply(void *map, int x, int y, int dx, int dy, void *src) {
  if (((LaserSimulator *)map)->onMap(x, y))
    ((LaserSimulator *)map)->setSeen(x, y);
}


/**
 * Function called by libfov to determine whether light can pass
 * through a cell. Return zero if light can pass though the cell at
 * (x,y), non-zero if it cannot.
 *
 * \param map Pointer to map data structure passed to function such as
 *            fov_circle.
 * \param x   Absolute x-axis position of cell.
 * \param y   Absolute x-axis position of cell.
 */
bool opaque(void *map, int x, int y) {
  return ((LaserSimulator *)map)->blockLOS(x, y);
}



/// - constructor --------------------------------------------------------------
LaserSimulator::LaserSimulator(vtkSmartPointer<vtkPNGReader> reader) 
{
  fov_settings_init(&fov_settings);
  fov_settings_set_shape(&fov_settings, FOV_SHAPE_CIRCLE);
  fov_settings_set_opacity_test_function(&fov_settings, opaque);
  fov_settings_set_apply_lighting_function(&fov_settings, apply);

  laserConf.count = 361;
  laserConf.maxRange = 30.0;
  laserConf.resolution = M_PI/360.0;
  laserConf.minAngle = -M_PI/2;
  laserConf.maxAngle = M_PI/2;      
  
  vtkSmartPointer<vtkImageData> imageData;
  imageData = reader->GetOutput();
  reader->Update();
  int dims[3];
  imageData->GetDimensions (dims);
//   std::cout << "DIMS: " << dims[0] << " " << dims[1] << " " << dims[2] << std::endl;
  grid = new ByteMatrix(dims[1], dims[0]);
  for(size_t r=0;r<dims[1];r++) {
    for(size_t c=0;c<dims[0];c++) {
      unsigned char val = *static_cast<unsigned char*>(imageData->GetScalarPointer(c, r, 0));
      (*grid)(r,c) = (val==0 ? OCCUPIED_CELL : (val==255 ? FREESPACE_CELL : UNKNOWN_CELL));
      //std::cout << (int) val << " ";
    }
  }
  W = dims[0];
  H = dims[1];
  
  
  K = 51.03;
  x0 = -16.96;
  y0 = -43.25;
  cellSize = 1.0/K;
 
  ///  maxRange = cfg.get<double>("laser-max-range");
  ///  cellSize = cfg.get<double>("grid-cell-size");
}

/// - destructor --------------------------------------------------------------
LaserSimulator::~LaserSimulator() {
  fov_settings_free(&fov_settings);
}



/// - public --------------------------------------------------------------
bool LaserSimulator::onMap(unsigned int x, unsigned int y) {
  return (x<W && y<H);
}


/// - public --------------------------------------------------------------
bool LaserSimulator::blockLOS(unsigned int x, unsigned int y) {
  if(!onMap(x,y)) return true;
  return ((*grid)(y, x) == OCCUPIED_CELL); //!=FREESPACE_CELL
}

/// - private --------------------------------------------------------------
// dir is in radians
int LaserSimulator::getScanIndex(double dir, bool down) {
  static const double density = double(numSamples)/(2*M_PI); // number of samples per degree
  double tmp = density*dir;
//  int result = down ? ceil(tmp) : floor(tmp);
  int result = down ? floor(tmp) : floor(tmp);
  if (result<0) { result+=numSamples; };
  if (result>=numSamples) { result-=numSamples; };
  return result;
}

/// - private --------------------------------------------------------------
double LaserSimulator::getAngle(double x, double y) {
  const double diagCellSize = cellSize*sqrt(2);
  double dx = grid2realX(x) - robot_x;
  double dy = grid2realY(y) - robot_y;
  double dist = sqrt(dx*dx + dy*dy)+2.0*cellSize;
  double angle = atan2(dy,dx);
  double delta = diagCellSize/(2*dist);
  int start = getScanIndex(angle-delta,true);
  int end = getScanIndex(angle+delta,false);
  
  
  
  
  //   DEBUG("obst: " << x << " " << y << " robot: " << map->real2gridX(robot_x) << " " << map->real2gridY(robot_y));
  //   DEBUG("obst: " << map->grid2realX(x) << " " << map->grid2realY(y) << " robot: " << robot_x << " " << robot_y);
//   std::cout << "dist: " << dist << " " << angle << " " << delta << "    " << start << " " << end << "   " << scan[start] << std::endl;
  //   DEBUG(angle-delta << " " << angle+ delta);
  //   DEBUG("start: " << start << " " << end);
//  points.push_back(Point(grid2realX(x),grid2realY(y)));
  if (start <= end) {
    for(int i=start;i<=end;i++) {
      if (scan[i] > dist) {
        scan[i] = dist;
      }
    }
  } else {
    for(int i=start;i<numSamples;i++) {
      if (scan[i] > dist) {
        scan[i] = dist;
      }
    }
    for(int i=0;i<=end;i++) {
      if (scan[i] > dist) {
        scan[i] = dist;
      }
    }
  }
  
  return angle;
}

/// - public --------------------------------------------------------------
void LaserSimulator::setSeen(unsigned int x, unsigned int y) {
  if(!onMap(x,y)) return;
  if ((*grid)(y,x) != FREESPACE_CELL) {
//     std::cout << "[" << x << " " << y << "] " << (int) (*grid)(y,x) << "]" << std::endl;
    getAngle(x, y);
  }
}

/// - public --------------------------------------------------------------
LaserScan LaserSimulator::getScan(const RobotPosition &pose) {
  LaserScan result;
  numSamples = ceil(2*M_PI/laserConf.resolution);
  points.resize(laserConf.count);
//  points.clear();
  result.resize(laserConf.count);
  //   DEBUG("laser: count " << lconf.count);
  //   DEBUG("laser: maxRange " << lconf.maxRange);
  //   DEBUG("laser: resolution " << lconf.resolution);
  //   DEBUG("laser: minAngle " << lconf.minAngle);
  //   DEBUG("laser: maxAngle " << lconf.maxAngle);
  //   DEBUG("laser: numSamples " << numSamples);
  
  robot_x = pose.x;
  robot_y = pose.y;
  unsigned int radius = double(laserConf.maxRange)/cellSize;
  scan.resize(numSamples);
  
  for(int i=0;i<numSamples;i++) {
    scan[i] = laserConf.maxRange;
  }
  
  fov_circle(&fov_settings, this, NULL, real2gridX(pose.x), real2gridY(pose.y), 2000);
  
  double x, y;
  int start = getScanIndex(pose.phi + laserConf.minAngle,true);
  int end  = getScanIndex(pose.phi + laserConf.maxAngle,false);
  //  std::stringstream ss;
  double a=pose.phi + laserConf.minAngle;
//   std::cout << "Pose " << pose.x << " " << pose.y << "    " << radius << std::endl;
//   std::cout << "Grid pose " << real2gridX(pose.x) << " " << real2gridY(pose.y) << std::endl;

  for(int n = 0, i=start;n<laserConf.count;a+=laserConf.resolution,n++) {
    //    ss << "n: " << n << " " << i << " " << scan[i] << " ";
    result[n] = scan[i];
    points[n].x = pose.x + result[n]*cos(a);
    points[n].y = pose.y + result[n]*sin(a);
//     std::cout << points[n].x << " " << points[n].y << "  " << result[n] << std::endl;
    
//     if (draw) {
//       gui->drawPoint(map->real2gridY(y),map->real2gridX(x),0, 50, 250);
//     }
    i = i==numSamples-1 ? 0 : i+1;    
  }
  //  DEBUG(ss.str());
  return result;
}

RawPoints LaserSimulator::getRawPoints() {
  return points;
}

bool LaserSimulator::isFeasible(RobotPosition pos) {
  return (*grid)(real2gridY(pos.y), real2gridX(pos.x)) == FREESPACE_CELL; 
}

/* end of sndsimulator.cc */
